@@ -1,8 +1,6 @@
-/* $XFree86: xc/programs/Xserver/hw/xfree86/drivers/fbdev/fbdev.c,v 1.44 2003/09/24 02:43:21 dawes Exp $ */
-
 /*
  * Authors:  Alan Hourihane, <alanh@fairlite.demon.co.uk>
- *	     Michel Dänzer, <michdaen@iiic.ethz.ch>
+ *	     Michel Dänzer, <michel@tungstengraphics.com>
  */
 
 #ifdef HAVE_CONFIG_H
@@ -29,31 +27,38 @@
 #include "afb.h"
 #endif
 
+#if GET_ABI_MAJOR(ABI_VIDEODRV_VERSION) < 6
 #include "xf86Resources.h"
 #include "xf86RAC.h"
+#endif
 
 #include "fbdevhw.h"
 
 #include "xf86xv.h"
 
-#define DEBUG 0
-
-#if DEBUG
-# define TRACE_ENTER(str)       ErrorF("fbdev: " str " %d\n",pScrn->scrnIndex)
-# define TRACE_EXIT(str)        ErrorF("fbdev: " str " done\n")
-# define TRACE(str)             ErrorF("fbdev trace: " str "\n")
-#else
-# define TRACE_ENTER(str)
-# define TRACE_EXIT(str)
-# define TRACE(str)
+#ifdef XSERVER_LIBPCIACCESS
+#include <pciaccess.h>
 #endif
 
+static Bool debug = 0;
+
+#define TRACE_ENTER(str) \
+    do { if (debug) ErrorF("fbdev: " str " %d\n",pScrn->scrnIndex); } while (0)
+#define TRACE_EXIT(str) \
+    do { if (debug) ErrorF("fbdev: " str " done\n"); } while (0)
+#define TRACE(str) \
+    do { if (debug) ErrorF("fbdev trace: " str "\n"); } while (0)
+
 /* -------------------------------------------------------------------- */
 /* prototypes                                                           */
 
 static const OptionInfoRec * FBDevAvailableOptions(int chipid, int busid);
 static void	FBDevIdentify(int flags);
 static Bool	FBDevProbe(DriverPtr drv, int flags);
+#ifdef XSERVER_LIBPCIACCESS
+static Bool	FBDevPciProbe(DriverPtr drv, int entity_num,
+     struct pci_device *dev, intptr_t match_data);
+#endif
 static Bool	FBDevPreInit(ScrnInfoPtr pScrn, int flags);
 static Bool	FBDevScreenInit(int Index, ScreenPtr pScreen, int argc,
 				char **argv);
@@ -68,11 +73,6 @@ static Bool	FBDevDriverFunc(ScrnInfoPtr pScrn, xorgDriverFuncOp op,
 
 enum { FBDEV_ROTATE_NONE=0, FBDEV_ROTATE_CW=270, FBDEV_ROTATE_UD=180, FBDEV_ROTATE_CCW=90 };
 
-/*static ShadowUpdateProc updateFuncs[] =
-  { shadowUpdatePacked, shadowUpdateRotate8_270, shadowUpdateRotate8_180, shadowUpdateRotate8_90,
-    shadowUpdatePacked, shadowUpdateRotate16_270, shadowUpdateRotate16_180, shadowUpdateRotate16_90,
-    shadowUpdatePacked, shadowUpdateRotate32_270, shadowUpdateRotate32_180, shadowUpdateRotate32_90 }; */
-
 
 /* -------------------------------------------------------------------- */
 
@@ -85,8 +85,17 @@ static int pix24bpp = 0;
 #define FBDEV_VERSION		4000
 #define FBDEV_NAME		"FBDEV"
 #define FBDEV_DRIVER_NAME	"fbdev"
-#define FBDEV_MAJOR_VERSION	0
-#define FBDEV_MINOR_VERSION	3
+
+#ifdef XSERVER_LIBPCIACCESS
+static const struct pci_id_match fbdev_device_match[] = {
+    {
+	PCI_MATCH_ANY, PCI_MATCH_ANY, PCI_MATCH_ANY, PCI_MATCH_ANY,
+	0x00030000, 0x00ffffff, 0
+    },
+
+    { 0, 0, 0 },
+};
+#endif
 
 _X_EXPORT DriverRec FBDEV = {
 	FBDEV_VERSION,
@@ -99,7 +108,12 @@ _X_EXPORT DriverRec FBDEV = {
 	FBDevAvailableOptions,
 	NULL,
 	0,
-	FBDevDriverFunc
+	FBDevDriverFunc,
+
+#ifdef XSERVER_LIBPCIACCESS
+    fbdev_device_match,
+    FBDevPciProbe
+#endif
 };
 
 /* Supported "chipsets" */
@@ -115,79 +129,20 @@ static SymTabRec FBDevChipsets[] = {
 typedef enum {
 	OPTION_SHADOW_FB,
 	OPTION_ROTATE,
-	OPTION_FBDEV
+	OPTION_FBDEV,
+	OPTION_DEBUG
 } FBDevOpts;
 
 static const OptionInfoRec FBDevOptions[] = {
 	{ OPTION_SHADOW_FB,	"ShadowFB",	OPTV_BOOLEAN,	{0},	FALSE },
 	{ OPTION_ROTATE,	"Rotate",	OPTV_STRING,	{0},	FALSE },
 	{ OPTION_FBDEV,		"fbdev",	OPTV_STRING,	{0},	FALSE },
+	{ OPTION_DEBUG,		"debug",	OPTV_BOOLEAN,	{0},	FALSE },
 	{ -1,			NULL,		OPTV_NONE,	{0},	FALSE }
 };
 
 /* -------------------------------------------------------------------- */
 
-static const char *afbSymbols[] = {
-	"afbScreenInit",
-	"afbCreateDefColormap",
-	NULL
-};
-
-static const char *fbSymbols[] = {
-	"fbScreenInit",
-	"fbPictureInit",
-	NULL
-};
-
-static const char *shadowSymbols[] = {
-	"shadowAdd",
-	"shadowInit",
-	"shadowSetup",
-	"shadowUpdatePacked",
-	"shadowUpdatePackedWeak",
-	"shadowUpdateRotatePacked",
-	"shadowUpdateRotatePackedWeak",
-	NULL
-};
-
-static const char *fbdevHWSymbols[] = {
-	"fbdevHWInit",
-	"fbdevHWProbe",
-	"fbdevHWSetVideoModes",
-	"fbdevHWUseBuildinMode",
-
-	"fbdevHWGetDepth",
-	"fbdevHWGetLineLength",
-	"fbdevHWGetName",
-	"fbdevHWGetType",
-	"fbdevHWGetVidmem",
-	"fbdevHWLinearOffset",
-	"fbdevHWLoadPalette",
-	"fbdevHWMapVidmem",
-	"fbdevHWUnmapVidmem",
-
-	/* colormap */
-	"fbdevHWLoadPalette",
-	"fbdevHWLoadPaletteWeak",
-
-	/* ScrnInfo hooks */
-	"fbdevHWAdjustFrameWeak",
-	"fbdevHWEnterVTWeak",
-	"fbdevHWLeaveVTWeak",
-	"fbdevHWModeInit",
-	"fbdevHWRestore",
-	"fbdevHWSave",
-	"fbdevHWSaveScreen",
-	"fbdevHWSaveScreenWeak",
-	"fbdevHWSwitchModeWeak",
-	"fbdevHWValidModeWeak",
-
-	"fbdevHWDPMSSet",
-	"fbdevHWDPMSSetWeak",
-
-	NULL
-};
-
 #ifdef XFree86LOADER
 
 MODULESETUPPROTO(FBDevSetup);
@@ -199,7 +154,7 @@ static XF86ModuleVersionInfo FBDevVersRec =
 	MODINFOSTRING1,
 	MODINFOSTRING2,
 	XORG_VERSION_CURRENT,
-	FBDEV_MAJOR_VERSION, FBDEV_MINOR_VERSION, 0,
+	PACKAGE_VERSION_MAJOR, PACKAGE_VERSION_MINOR, PACKAGE_VERSION_PATCHLEVEL,
 	ABI_CLASS_VIDEODRV,
 	ABI_VIDEODRV_VERSION,
 	NULL,
@@ -216,8 +171,6 @@ FBDevSetup(pointer module, pointer opts, int *errmaj, int *errmin)
 	if (!setupDone) {
 		setupDone = TRUE;
 		xf86AddDriver(&FBDEV, module, HaveDriverFuncs);
-		LoaderRefSymLists(afbSymbols, fbSymbols,
-				  shadowSymbols, fbdevHWSymbols, NULL);
 		return (pointer)1;
 	} else {
 		if (errmaj) *errmaj = LDR_ONCEONLY;
@@ -237,7 +190,9 @@ typedef struct {
 	int				lineLength;
 	int				rotate;
 	Bool				shadowFB;
+	void				*shadow;
 	CloseScreenProcPtr		CloseScreen;
+	CreateScreenResourcesProcPtr	CreateScreenResources;
 	void				(*PointerMoved)(int index, int x, int y);
 	EntityInfoPtr			pEnt;
 	/* DGA info */
@@ -281,6 +236,53 @@ FBDevIdentify(int flags)
 	xf86PrintChipsets(FBDEV_NAME, "driver for framebuffer", FBDevChipsets);
 }
 
+
+#ifdef XSERVER_LIBPCIACCESS
+static Bool FBDevPciProbe(DriverPtr drv, int entity_num,
+			  struct pci_device *dev, intptr_t match_data)
+{
+    ScrnInfoPtr pScrn = NULL;
+
+    if (!xf86LoadDrvSubModule(drv, "fbdevhw"))
+	return FALSE;
+	    
+    pScrn = xf86ConfigPciEntity(NULL, 0, entity_num, NULL, NULL,
+				NULL, NULL, NULL, NULL);
+    if (pScrn) {
+	char *device;
+	GDevPtr devSection = xf86GetDevFromEntity(pScrn->entityList[0],
+						  pScrn->entityInstanceList[0]);
+
+	device = xf86FindOptionValue(devSection->options, "fbdev");
+	if (fbdevHWProbe(NULL, device, NULL)) {
+	    pScrn->driverVersion = FBDEV_VERSION;
+	    pScrn->driverName    = FBDEV_DRIVER_NAME;
+	    pScrn->name          = FBDEV_NAME;
+	    pScrn->Probe         = FBDevProbe;
+	    pScrn->PreInit       = FBDevPreInit;
+	    pScrn->ScreenInit    = FBDevScreenInit;
+	    pScrn->SwitchMode    = fbdevHWSwitchModeWeak();
+	    pScrn->AdjustFrame   = fbdevHWAdjustFrameWeak();
+	    pScrn->EnterVT       = fbdevHWEnterVTWeak();
+	    pScrn->LeaveVT       = fbdevHWLeaveVTWeak();
+	    pScrn->ValidMode     = fbdevHWValidModeWeak();
+
+	    xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
+		       "claimed PCI slot %d@%d:%d:%d\n", 
+		       dev->bus, dev->domain, dev->dev, dev->func);
+	    xf86DrvMsg(pScrn->scrnIndex, X_INFO,
+		       "using %s\n", device ? device : "default device");
+	}
+	else {
+	    pScrn = NULL;
+	}
+    }
+
+    return (pScrn != NULL);
+}
+#endif
+
+
 static Bool
 FBDevProbe(DriverPtr drv, int flags)
 {
@@ -304,26 +306,32 @@ FBDevProbe(DriverPtr drv, int flags)
 	if (!xf86LoadDrvSubModule(drv, "fbdevhw"))
 	    return FALSE;
 	    
-	xf86LoaderReqSymLists(fbdevHWSymbols, NULL);
-	
 	for (i = 0; i < numDevSections; i++) {
 	    Bool isIsa = FALSE;
 	    Bool isPci = FALSE;
 
 	    dev = xf86FindOptionValue(devSections[i]->options,"fbdev");
 	    if (devSections[i]->busID) {
+#ifndef XSERVER_LIBPCIACCESS
 	        if (xf86ParsePciBusString(devSections[i]->busID,&bus,&device,
 					  &func)) {
 		    if (!xf86CheckPciSlot(bus,device,func))
 		        continue;
 		    isPci = TRUE;
-		} else if (xf86ParseIsaBusString(devSections[i]->busID))
+		} else
+#endif
+#ifdef HAVE_ISA
+		if (xf86ParseIsaBusString(devSections[i]->busID))
 		    isIsa = TRUE;
+		else
+#endif
+		    0;
 		  
 	    }
 	    if (fbdevHWProbe(NULL,dev,NULL)) {
 		pScrn = NULL;
 		if (isPci) {
+#ifndef XSERVER_LIBPCIACCESS
 		    /* XXX what about when there's no busID set? */
 		    int entity;
 		    
@@ -339,7 +347,9 @@ FBDevProbe(DriverPtr drv, int flags)
 		    xf86DrvMsg(pScrn->scrnIndex, X_CONFIG,
 			       "claimed PCI slot %d:%d:%d\n",bus,device,func);
 
+#endif
 		} else if (isIsa) {
+#ifdef HAVE_ISA
 		    int entity;
 		    
 		    entity = xf86ClaimIsaSlot(drv, 0,
@@ -347,6 +357,7 @@ FBDevProbe(DriverPtr drv, int flags)
 		    pScrn = xf86ConfigIsaEntity(pScrn,0,entity,
 						      NULL,RES_SHARED_VGA,
 						      NULL,NULL,NULL,NULL);
+#endif
 		} else {
 		   int entity;
 
@@ -387,7 +398,6 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 	FBDevPtr fPtr;
 	int default_depth, fbbpp;
 	const char *mod = NULL, *s;
-	const char **syms = NULL;
 	int type;
 
 	if (flags & PROBE_DETECT) return FALSE;
@@ -405,6 +415,7 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 
 	fPtr->pEnt = xf86GetEntityInfo(pScrn->entityList[0]);
 
+#ifndef XSERVER_LIBPCIACCESS
 	pScrn->racMemFlags = RAC_FB | RAC_COLORMAP | RAC_CURSOR | RAC_VIEWPORT;
 	/* XXX Is this right?  Can probably remove RAC_FB */
 	pScrn->racIoFlags = RAC_FB | RAC_COLORMAP | RAC_CURSOR | RAC_VIEWPORT;
@@ -415,13 +426,13 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 		   "xf86RegisterResources() found resource conflicts\n");
 		return FALSE;
 	}
-
+#endif
 	/* open device */
 	if (!fbdevHWInit(pScrn,NULL,xf86FindOptionValue(fPtr->pEnt->device->options,"fbdev")))
 		return FALSE;
 	default_depth = fbdevHWGetDepth(pScrn,&fbbpp);
 	if (!xf86SetDepthBpp(pScrn, default_depth, default_depth, fbbpp,
-			     Support24bppFb | Support32bppFb | SupportConvert32to24 | PreferConvert32to24))
+			     Support24bppFb | Support32bppFb | SupportConvert32to24 | SupportConvert24to32))
 		return FALSE;
 	xf86PrintDepthBpp(pScrn);
 
@@ -474,6 +485,8 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 	/* use shadow framebuffer by default */
 	fPtr->shadowFB = xf86ReturnOptValBool(fPtr->Options, OPTION_SHADOW_FB, TRUE);
 
+	debug = xf86ReturnOptValBool(fPtr->Options, OPTION_DEBUG, FALSE);
+
 	/* rotation */
 	fPtr->rotate = FBDEV_ROTATE_NONE;
 	if ((s = xf86GetOptValString(fPtr->Options, OPTION_ROTATE)))
@@ -529,15 +542,8 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 		fbdevHWUseBuildinMode(pScrn);
 	pScrn->currentMode = pScrn->modes;
 
-	if (fPtr->shadowFB)
-		pScrn->displayWidth = pScrn->virtualX;	/* ShadowFB handles this correctly */
-	else {
-		int fbbpp;
-		/* FIXME: this doesn't work for all cases, e.g. when each scanline
-			has a padding which is independent from the depth (controlfb) */
-		fbdevHWGetDepth(pScrn,&fbbpp);
-		pScrn->displayWidth = fbdevHWGetLineLength(pScrn)/(fbbpp >> 3);
-	}
+	/* First approximation, may be refined in ScreenInit */
+	pScrn->displayWidth = pScrn->virtualX;
 
 	xf86PrintModes(pScrn);
 
@@ -549,7 +555,6 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 	{
 	case FBDEVHW_PLANES:
 		mod = "afb";
-		syms = afbSymbols;
 		break;
 	case FBDEVHW_PACKED_PIXELS:
 		switch (pScrn->bitsPerPixel)
@@ -559,7 +564,6 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 		case 24:
 		case 32:
 			mod = "fb";
-			syms = fbSymbols;
 			break;
 		default:
 			xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
@@ -595,9 +599,6 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 		FBDevFreeRec(pScrn);
 		return FALSE;
 	}
-	if (mod && syms) {
-		xf86LoaderReqSymLists(syms, NULL);
-	}
 
 	/* Load shadow if needed */
 	if (fPtr->shadowFB) {
@@ -607,40 +608,53 @@ FBDevPreInit(ScrnInfoPtr pScrn, int flags)
 			FBDevFreeRec(pScrn);
 			return FALSE;
 		}
-		xf86LoaderReqSymLists(shadowSymbols, NULL);
 	}
 
 	TRACE_EXIT("PreInit");
 	return TRUE;
 }
 
+
 static Bool
-FBDevShadowInit(ScreenPtr pScreen, FBDevPtr fPtr)
+FBDevCreateScreenResources(ScreenPtr pScreen)
 {
     PixmapPtr pPixmap;
-    ShadowUpdateProc update;
-    ShadowWindowProc window;
-    
-    pPixmap = pScreen->CreatePixmap(pScreen, pScreen->width, pScreen->height,
-				    pScreen->rootDepth);
-    if (!pPixmap)
+    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
+    FBDevPtr fPtr = FBDEVPTR(pScrn);
+    Bool ret;
+
+    pScreen->CreateScreenResources = fPtr->CreateScreenResources;
+    ret = pScreen->CreateScreenResources(pScreen);
+    pScreen->CreateScreenResources = FBDevCreateScreenResources;
+
+    if (!ret)
 	return FALSE;
-    
-    if (!shadowSetup(pScreen)) {
-	pScreen->DestroyPixmap(pPixmap);
+
+    pPixmap = pScreen->GetScreenPixmap(pScreen);
+
+    if (!shadowAdd(pScreen, pPixmap, fPtr->rotate ?
+		   shadowUpdateRotatePackedWeak() : shadowUpdatePackedWeak(),
+		   FBDevWindowLinear, fPtr->rotate, NULL)) {
 	return FALSE;
     }
 
-    update = fPtr->rotate ? shadowUpdateRotatePackedWeak()
-	: shadowUpdatePackedWeak();
+    return TRUE;
+}
 
-    if (!shadowAdd(pScreen, pPixmap, update, FBDevWindowLinear,
-		   fPtr->rotate, NULL)) {
-	pScreen->DestroyPixmap(pPixmap);
+static Bool
+FBDevShadowInit(ScreenPtr pScreen)
+{
+    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
+    FBDevPtr fPtr = FBDEVPTR(pScrn);
+    
+    if (!shadowSetup(pScreen)) {
 	return FALSE;
-    } else {
-	return TRUE;
-    }	    
+    }
+
+    fPtr->CreateScreenResources = pScreen->CreateScreenResources;
+    pScreen->CreateScreenResources = FBDevCreateScreenResources;
+
+    return TRUE;
 }
 
 
@@ -651,7 +665,7 @@ FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 	FBDevPtr fPtr = FBDEVPTR(pScrn);
 	VisualPtr visual;
 	int init_picture = 0;
-	int ret,flags,width,height;
+	int ret, flags;
 	int type;
 
 	TRACE_ENTER("FBDevScreenInit");
@@ -708,11 +722,20 @@ FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 
 	if(fPtr->rotate==FBDEV_ROTATE_CW || fPtr->rotate==FBDEV_ROTATE_CCW)
 	{
-	  height = pScrn->virtualX;
-	  width = pScrn->displayWidth = pScrn->virtualY;
-	} else {
-	  height = pScrn->virtualY;
-	  width = pScrn->virtualX;
+	  int tmp = pScrn->virtualX;
+	  pScrn->virtualX = pScrn->displayWidth = pScrn->virtualY;
+	  pScrn->virtualY = tmp;
+	} else if (!fPtr->shadowFB) {
+		/* FIXME: this doesn't work for all cases, e.g. when each scanline
+			has a padding which is independent from the depth (controlfb) */
+		pScrn->displayWidth = fbdevHWGetLineLength(pScrn) /
+				      (pScrn->bitsPerPixel / 8);
+
+		if (pScrn->displayWidth != pScrn->virtualX) {
+			xf86DrvMsg(scrnIndex, X_INFO,
+				   "Pitch updated to %d after ModeInit\n",
+				   pScrn->displayWidth);
+		}
 	}
 
 	if(fPtr->rotate && !fPtr->PointerMoved) {
@@ -722,6 +745,17 @@ FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 
 	fPtr->fbstart = fPtr->fbmem + fPtr->fboff;
 
+	if (fPtr->shadowFB) {
+	    fPtr->shadow = xcalloc(1, pScrn->virtualX * pScrn->virtualY *
+				   pScrn->bitsPerPixel);
+
+	    if (!fPtr->shadow) {
+		xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
+			   "Failed to allocate shadow framebuffer\n");
+		return FALSE;
+	    }
+	}
+
 	switch ((type = fbdevHWGetType(pScrn)))
 	{
 #ifdef USE_AFB
@@ -752,8 +786,11 @@ FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 		case 16:
 		case 24:
 		case 32:
-			ret = fbScreenInit(pScreen, fPtr->fbstart, width, height,
-					   pScrn->xDpi, pScrn->yDpi, pScrn->displayWidth, pScrn->bitsPerPixel);
+			ret = fbScreenInit(pScreen, fPtr->shadowFB ? fPtr->shadow
+					   : fPtr->fbstart, pScrn->virtualX,
+					   pScrn->virtualY, pScrn->xDpi,
+					   pScrn->yDpi, pScrn->displayWidth,
+					   pScrn->bitsPerPixel);
 			init_picture = 1;
 			break;
 	 	default:
@@ -791,7 +828,7 @@ FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 	default:
 		xf86DrvMsg(scrnIndex, X_ERROR,
 		           "internal error: unrecognised hardware type (%d) "
-			   "encountered in FBDevScreenInit()\n");
+			   "encountered in FBDevScreenInit()\n", type);
 		ret = FALSE;
 		break;
 	}
@@ -818,7 +855,7 @@ FBDevScreenInit(int scrnIndex, ScreenPtr pScreen, int argc, char **argv)
 		xf86DrvMsg(pScrn->scrnIndex, X_WARNING,
 			   "Render extension initialisation failed\n");
 
-	if (fPtr->shadowFB && !FBDevShadowInit(pScreen, fPtr)) {
+	if (fPtr->shadowFB && !FBDevShadowInit(pScreen)) {
 	    xf86DrvMsg(scrnIndex, X_ERROR,
 		       "shadow framebuffer initialization failed\n");
 	    return FALSE;
@@ -921,6 +958,10 @@ FBDevCloseScreen(int scrnIndex, ScreenPtr pScreen)
 	
 	fbdevHWRestore(pScrn);
 	fbdevHWUnmapVidmem(pScrn);
+	if (fPtr->shadow) {
+	    xfree(fPtr->shadow);
+	    fPtr->shadow = NULL;
+	}
 	if (fPtr->pDGAMode) {
 	  xfree(fPtr->pDGAMode);
 	  fPtr->pDGAMode = NULL;
@@ -928,6 +969,7 @@ FBDevCloseScreen(int scrnIndex, ScreenPtr pScreen)
 	}
 	pScrn->vtSema = FALSE;
 
+	pScreen->CreateScreenResources = fPtr->CreateScreenResources;
 	pScreen->CloseScreen = fPtr->CloseScreen;
 	return (*pScreen->CloseScreen)(scrnIndex, pScreen);
 }
@@ -953,7 +995,7 @@ FBDevWindowLinear(ScreenPtr pScreen, CARD32 row, CARD32 offset, int mode,
     else
       *size = fPtr->lineLength = fbdevHWGetLineLength(pScrn);
 
-    return ((CARD8 *)fPtr->fbmem + fPtr->fboff + row * fPtr->lineLength + offset);
+    return ((CARD8 *)fPtr->fbstart + row * fPtr->lineLength + offset);
 }
 
 static void
@@ -1126,6 +1168,7 @@ FBDevDGAAddModes(ScrnInfoPtr pScrn)
 static Bool
 FBDevDGAInit(ScrnInfoPtr pScrn, ScreenPtr pScreen)
 {
+#ifdef XFreeXDGA
     FBDevPtr fPtr = FBDEVPTR(pScrn);
 
     if (pScrn->depth < 8)
@@ -1136,6 +1179,9 @@ FBDevDGAInit(ScrnInfoPtr pScrn, ScreenPtr pScreen)
 
     return (DGAInit(pScreen, &FBDevDGAFunctions,
 	    fPtr->pDGAMode, fPtr->nDGAMode));
+#else
+    return TRUE;
+#endif
 }
 
 static Bool
